from enum import Enum, StrEnum

from .common import Item


class InPlageHoraire(Item):
    def _translate(self, state: str):
        return state == "true"


class InputAnalog(Item):
    def _translate(self, state: str):
        return float(state)


class InputString(Item):
    pass


class InputSwitch(Item):
    def _translate(self, state: str):
        return state == "true"


class InputSwitchLongPressState(Enum):
    NONE = 0
    SHORT = 1
    LONG = 2


class InputSwitchLongPress(Item):
    def _translate(self, state: str):
        return InputSwitchLongPressState(int(state))


class InputSwitchTripleState(Enum):
    NONE = 0
    SINGLE = 1
    DOUBLE = 2
    TRIPLE = 3


class InputSwitchTriple(Item):
    def _translate(self, state: str):
        return InputSwitchTripleState(int(state))


class InputTemp(Item):
    def _translate(self, state: str):
        return float(state)


class InputTime(Item):
    def _translate(self, state: str):
        return state == "true"


class InputTimer(Item):
    def _translate(self, state: str):
        return state == "true"

    def start(self):
        self._send("start")
        self._update_state()

    def stop(self):
        self._send("stop")
        self._update_state()

    def reset(self, hours, minutes, seconds, milliseconds):
        self._send(f"{hours}:{minutes}:{seconds}:{milliseconds}")
        self._update_state()


class InternalBool(Item):
    def _translate(self, state: str):
        return state == "true"

    def true(self):
        self._send("true")
        self._state = True

    def false(self):
        self._send("false")
        self._state = False

    def toggle(self):
        self._send("toggle")
        self._update_state()

    def impulse(self, *pattern):
        cmd = "impulse"
        for step in pattern:
            cmd += f" {step}"
        self._send(cmd)
        print(cmd)
        self._update_state()


class InternalInt(Item):
    def _translate(self, state: str):
        return int(state)

    def set(self, value):
        self._send(f"{value}")
        self._state = value

    def inc(self, value=0):
        if value == 0:
            self._send("inc")
        else:
            self._send(f"inc {value}")
        self._update_state()

    def dec(self, value=0):
        if value == 0:
            self._send("dec")
        else:
            self._send(f"dec {value}")
        self._update_state()


class InternalString(Item):
    def set(self, value):
        self._send(value)
        self._state = value


class OutputLight(Item):
    def _translate(self, state: str):
        return state == "true"

    def true(self):
        self._send("true")
        self._state = True

    def false(self):
        self._send("false")
        self._state = False

    def toggle(self):
        self._send("toggle")
        self._update_state()

    def impulse(self, *pattern):
        cmd = "impulse"
        for step in pattern:
            cmd += f" {step}"
        self._send(cmd)
        print(cmd)
        self._update_state()


class OutputLightDimmer(Item):
    def _translate(self, state: str):
        value = int(state)
        if value > 100:
            value = 100
        elif value < 0:
            value = 0
        return value

    def true(self):
        self._send("true")
        self._update_state()

    def false(self):
        self._send("false")
        self._state = 0

    def toggle(self):
        self._send("toggle")
        self._update_state()

    def impulse(self, *pattern):
        cmd = "impulse"
        for step in pattern:
            cmd += f" {step}"
        self._send(cmd)
        print(cmd)
        self._update_state()

    def set_off(self, value):
        if value < 1:
            value = 1
        elif value > 100:
            value = 100
        self._send(f"set off {value}")
        if self._state != 0:
            self._state = value

    def set(self, value):
        if value < 1:
            value = 1
        elif value > 100:
            value = 100
        self._send(f"set {value}")
        self._state = value

    def up(self, value):
        if value < 1:
            value = 1
        elif value > 100:
            value = 100
        self._send(f"up {value}")
        self._update_state()

    def down(self, value):
        if value < 1:
            value = 1
        elif value > 100:
            value = 100
        self._send(f"down {value}")
        self._update_state()

    def hold_press(self):
        self._send("hold press")

    def hold_stop(self):
        self._send("hold stop")
        self._update_state()


class OutputShutterAction(StrEnum):
    STATIONARY = ""
    UP = "up"
    DOWN = "down"
    STOP = "stop"
    CALIBRATION = "calibrate"


class OutputShutterSmart(Item):
    def _translate(self, state: str):
        infos = state.split()
        return {"action": OutputShutterAction(infos[0]), "position": int(infos[1])}

    def up(self):
        self._send("up")

    def down(self):
        self._send("down")

    def stop(self):
        self._send("stop")
        self._update_state()

    def toggle(self):
        self._send("toggle")
        self._update_state()

    def impulse_up(self, duration):
        self._send(f"impulse up {duration}")

    def impulse_down(self, duration):
        self._send(f"impulse down {duration}")

    def set(self, value):
        if value < 1:
            value = 1
        elif value > 100:
            value = 100
        self._send(f"set {value}")
        self._state = value

    def up(self, value):
        if value < 1:
            value = 1
        elif value > 100:
            value = 100
        self._send(f"up {value}")
        self._update_state()

    def down(self, value):
        if value < 1:
            value = 1
        elif value > 100:
            value = 100
        self._send(f"down {value}")
        self._update_state()

    def calibrate(self):
        self._send(f"calibrate")


class Scenario(Item):
    def _translate(self, state: str):
        return state == "true"

    def true(self):
        self._send("true")
        self._state = True

    def false(self):
        self._send("false")
        self._state = False



import logging

_LOGGER = logging.getLogger(__name__)


class Event:
    """An item state change event"""

    def __init__(self, item):
        """Initialize the event with the item and the current item state

        Create the event after changing the item state.
        """
        self._item = item
        self._state = item.state

    def __repr__(self):
        return f"{self._item.info} => {self._state}"

    def __eq__(self, other):
        if other == None:
            return False
        return self._item == other.item and self._state == other.state

    @property
    def item(self):
        """Item this event is related to"""
        return self._item

    @property
    def state(self):
        """State for this event"""
        return self._state


class Item:
    """An item in the Calaos installation"""

    _state = None

    def __init__(self, data, room, conn):
        self._id = data["id"]
        self._gui_type = data["gui_type"]
        self._io_type = data["io_type"]
        self._name = data["name"]
        self._type = data["type"]
        self._var_type = data["var_type"]
        self._visible = data["visible"] == "true"
        self.internal_set_state(data["state"])
        self._room = room
        self._conn = conn

    def __repr__(self):
        return f"{self.info} = {self.state}"

    def internal_set_state(self, state):
        self._state = self._translate(state)

    def internal_from_event(self, state):
        translated = self._translate(state)
        if translated == self._state:
            return False
        self._state = translated
        return True

    @property
    def info(self):
        return f"[{self._id}] {self._name} ({self._type}/{self._io_type})"

    @property
    def id(self):
        return self._id

    @property
    def gui_type(self):
        return self._gui_type

    @property
    def io_type(self):
        return self._io_type

    @property
    def name(self):
        return self._name

    @property
    def state(self):
        return self._state

    @property
    def type(self):
        return self._type

    @property
    def var_type(self):
        return self._var_type

    @property
    def visible(self):
        return self._visible

    @property
    def room(self):
        return self._room

    def _send(self, value):
        _LOGGER.debug(f"Setting state of {self._id} ({self._name}) with value: {value}")
        response = self._conn.send(
            {
                "action": "set_state",
                "type": self._io_type,
                "id": self._id,
                "value": value,
            }
        )
        if not response["success"]:
            _LOGGER.error(
                f"Failed to set state of {self._id} ({self._name}) with value: {value}"
            )

    def _update_state(self):
        result = self._conn.send({"action": "get_state", "items": [self._id]})
        self._state = self._translate(result[self._id])

    def _translate(self, state: str):
        return state


class Default(Item):
    pass



import json
import logging
import ssl
import time
import urllib.request

from .item import new_item
from .item.common import Event

_LOGGER = logging.getLogger(__name__)

# Calaos deletes registered polling uuids after 5 minutes
POLLING_MAX_WAIT = 5 * 60


class Room:
    """A room in the Calaos configuration"""

    def __init__(self, name: str, type: str):
        """Initialize the room

        Parameters:
            name (str):
                Name of the room

            type (str):
                Type of the room
        """
        self._name = name
        self._type = type
        self._items = []

    def __repr__(self):
        return f"{self._name} ({self._type}): {len(self._items)} items"

    @property
    def name(self):
        return self._name

    @property
    def type(self):
        return self._type

    @property
    def items(self):
        return self._items

    def _addItem(self, item):
        """Add a new item in the room

        Parameters:
            item (pycalaos.item.Item):
                The item to add

        Return nothing
        """
        self._items.append(item)


class _Conn:
    def __init__(self, uri, username, password):
        self._uri = f"{uri}/api.php"
        self._username = username
        self._password = password
        self._context = ssl._create_unverified_context()

    def send(self, request):
        request["cn_user"] = self._username
        request["cn_pass"] = self._password
        req = urllib.request.Request(
            self._uri,
            data=json.dumps(request).encode("ascii"),
            headers={"Content-Type": "application/json"},
        )
        with urllib.request.urlopen(req, context=self._context) as response:
            return json.load(response)


class Client:
    """A Calaos client"""

    def __init__(self, uri: str, username: str, password: str):
        """Initialize the client and load the home configuration and state.

        Parameters:
            uri (str):
                URI of the Calaos server (usually, "http[s]://A.B.C.D")

            username (str):
                Username to connect to the Calaos server

            password (str):
                Password to connect to the Calaos server
        """
        self._conn = _Conn(uri, username, password)
        self._polling_id = None
        self._last_poll = 0
        self.reload_home()

    def __repr__(self):
        return f"Calaos Client with {len(self.rooms)} rooms"

    def reload_home(self):
        """Reload the complete home configuration, resetting rooms and items

        This could be necessary if the Calaos server is reconfigured with
        Calaos Installer and the client is not restarted).

        Return nothing
        """
        _LOGGER.debug("Getting the whole home")
        resp = self._conn.send({"action": "get_home"})
        rooms = []
        items = {}
        items_by_type = {}
        for roomData in resp["home"]:
            room = Room(roomData["name"], roomData["type"])
            for itemData in roomData["items"]:
                item = new_item(itemData, room, self._conn)
                items[item._id] = item
                try:
                    items_by_type[type(item)].append(item)
                except KeyError:
                    items_by_type[type(item)] = [item]
                room._addItem(item)
            rooms.append(room)
        self._rooms = rooms
        self._items = items
        self._items_by_type = items_by_type

    def update_all(self):
        """Check all states and return events

        Return events for states changes (list of pycalaos.item.common.Event)
        """
        _LOGGER.debug("Getting all states from known items")
        resp = self._conn.send(
            {"action": "get_state", "items": list(self.items.keys())}
        )
        events = []
        for kv in resp.items():
            changed = self.items[kv[0]].internal_set_state(kv[1])
            if changed:
                events.append(Event(self.items[kv[0]]))
        return events

    def poll(self):
        """Change items states and return all events since the last poll

        Return events for states changes (list of pycalaos.item.common.Event)
        """
        now = time.time()
        if now - self._last_poll > POLLING_MAX_WAIT:
            _LOGGER.debug("Registering to the polling")
            # If there is no existing poll queue, create a new one and
            # try to get new states for all items
            resp = self._conn.send({"action": "poll_listen", "type": "register"})
            self._polling_id = resp["uuid"]
            events = self.update_all()
        else:
            resp = self._conn.send(
                {"action": "poll_listen", "type": "get", "uuid": self._polling_id}
            )
            if len(resp["events"]) > 0:
                _LOGGER.debug(f"Raw events from polling: {resp['events']}")
            events = []
            for rawEvent in resp["events"]:
                try:
                    item = self.items[rawEvent["data"]["id"]]
                except KeyError:
                    continue
                changed = item.internal_from_event(rawEvent["data"]["state"])
                if changed:
                    events.append(Event(item))
        self._last_poll = now
        if len(events) > 0:
            _LOGGER.debug(f"Events: {events}")
        return events

    @property
    def rooms(self):
        return self._rooms

    @property
    def items(self):
        """Items referenced by their IDs (dict of str: pycalaos.item.Item)"""
        return self._items

    @property
    def item_types(self):
        """Complete list of item types currently in use"""
        return list(self._items_by_type.keys())

    def items_by_type(self, type):
        """Return only the items of the given type"""
        try:
            return self._items_by_type[type]
        except KeyError:
            return []
